@@ -39,6 +39,13 @@
 #include "unicode-compose-table.h"
 #include "version.h"
 
+#if defined(FCFT_ENABLE_SVG_NANOSVG)
+ #include "svg-backend-nanosvg.h"
+#endif
+#if defined(FCFT_ENABLE_SVG_LIBRSVG)
+ #include "3rd-party/freetype-rsvg/rsvg-port.h"
+#endif
+
 #define min(x, y) ((x) < (y) ? (x) : (y))
 #define max(x, y) ((x) > (y) ? (x) : (y))
 #define ALEN(v) (sizeof(v) / sizeof((v)[0]))
@@ -169,6 +176,9 @@ fcft_capabilities(void)
 #if defined(FCFT_HAVE_HARFBUZZ) && defined(FCFT_HAVE_UTF8PROC)
     ret |= FCFT_CAPABILITY_TEXT_RUN_SHAPING;
 #endif
+#if defined(FCFT_ENABLE_SVG_LIBRSVG) || defined(FCFT_ENABLE_SVG_NANOSVG)
+    ret |= FCFT_CAPABILITY_SVG;
+#endif
 
     return ret;
 }
@@ -198,6 +208,12 @@ fcft_init(enum fcft_log_colorize colorize, bool do_syslog,
         return false;
     }
 
+#if defined(FCFT_ENABLE_SVG_NANOSVG)
+    FT_Property_Set(ft_lib, "ot-svg", "svg-hooks", &nanosvg_hooks);
+#elif defined(FCFT_ENABLE_SVG_LIBRSVG)
+    FT_Property_Set(ft_lib, "ot-svg", "svg-hooks", &rsvg_hooks);
+#endif
+
     FcInit();
 
     /*
@@ -303,16 +319,27 @@ log_version_information(void)
 
     enum fcft_capabilities caps = fcft_capabilities();
 
-    static char caps_str[256];
+    char caps_str[256];
     snprintf(
         caps_str, sizeof(caps_str),
-        "%cgraphemes %cruns %cassertions",
+        "%cgraphemes %cruns %csvg%s %cassertions",
         caps & FCFT_CAPABILITY_GRAPHEME_SHAPING ? '+' : '-',
         caps & FCFT_CAPABILITY_TEXT_RUN_SHAPING ? '+' : '-',
+        caps & FCFT_CAPABILITY_SVG ? '+' : '-',
+#if defined(FCFT_ENABLE_SVG_LIBRSVG)
+        "(librsvg)",
+#elif defined(FCFT_ENABLE_SVG_NANOSVG)
+        "(nanosvg)",
+#else
+        "",
+#endif
         is_assertions_enabled() ? '+' : '-');
 
     LOG_INFO("fcft: %s %s", FCFT_VERSION, caps_str);
 
+    char deps_str[128];
+    int deps_idx = 0;
+
     {
         int raw_version = FcGetVersion();
 
@@ -321,15 +348,33 @@ log_version_information(void)
         const int minor = raw_version / 100; raw_version %= 100;
         const int patch = raw_version;
 
-        LOG_INFO("fontconfig: %d.%d.%d", major, minor, patch);
+        deps_idx += snprintf(
+            &deps_str[deps_idx], sizeof(deps_str) - deps_idx,
+            "fontconfig: %d.%d.%d", major, minor, patch);
     }
 
     {
         int major, minor, patch;
         FT_Library_Version(ft_lib, &major, &minor, &patch);
-        LOG_INFO("freetype: %d.%d.%d", major, minor, patch);
+        deps_idx += snprintf(
+            &deps_str[deps_idx], sizeof(deps_str) - deps_idx,
+            ", freetype: %d.%d.%d", major, minor, patch);
     }
 
+#if defined(FCFT_HAVE_HARFBUZZ)
+    deps_idx += snprintf(
+        &deps_str[deps_idx], sizeof(deps_str) - deps_idx,
+        ", harfbuzz: %s", hb_version_string());
+#endif
+
+#if defined(FCFT_HAVE_UTF8PROC)
+    deps_idx += snprintf(
+        &deps_str[deps_idx], sizeof(deps_str) - deps_idx,
+        ", utf8proc: %s (Unicode %s)",
+        utf8proc_version(), utf8proc_unicode_version());
+#endif
+
+    LOG_INFO("%.*s", deps_idx, deps_str);
 }
 
 FCFT_EXPORT bool
@@ -534,20 +579,20 @@ instantiate_pattern(FcPattern *pattern, double req_pt_size, double req_px_size,
     }
 
     FcChar8 *full_name = NULL;
-    if (FcPatternGetString(pattern, FC_FULLNAME, 0, &full_name) != FcResultMatch)
+    if (FcPatternGetString(pattern, FC_FULLNAME, face_index, &full_name) != FcResultMatch)
         LOG_WARN("failed to get full font name");
 
     mtx_lock(&ft_lock);
     FT_Face ft_face;
     FT_Error ft_err = FT_New_Face(ft_lib, (const char *)face_file, face_index, &ft_face);
     mtx_unlock(&ft_lock);
-    if (ft_err != 0) {
+    if (ft_err != FT_Err_Ok) {
         LOG_ERR("%s: failed to create FreeType face; %s",
                 face_file, ft_error_string(ft_err));
         return false;
     }
 
-    if ((ft_err = FT_Set_Pixel_Sizes(ft_face, 0, round(pixel_size))) != 0) {
+    if ((ft_err = FT_Set_Pixel_Sizes(ft_face, 0, round(pixel_size))) != FT_Err_Ok) {
         LOG_ERR("%s: failed to set character size: %s",
                 face_file, ft_error_string(ft_err));
         goto err_done_face;
@@ -785,11 +830,10 @@ instantiate_pattern(FcPattern *pattern, double req_pt_size, double req_px_size,
 #if defined(LOG_ENABLE_DBG) && LOG_ENABLE_DBG
     LOG_DBG("%s: size=%.2fpt/%.2fpx, dpi=%.2f, fixup-factor: %.4f, "
             "line-height: %dpx, ascent: %dpx, descent: %dpx, "
-            "x-advance (max/space): %d/%dpx, features:%s",
+            "x-advance (max): %dpx, features:%s",
             font->path, size, pixel_size, dpi, font->pixel_size_fixup,
             font->metrics.height, font->metrics.ascent, font->metrics.descent,
-            font->metrics.max_advance.x, font->metrics.space_advance.x,
-            features);
+            font->metrics.max_advance.x, features);
 #else
     LOG_INFO("%s: size=%.2fpt/%dpx, dpi=%.2f%s",
              font->path, size, (int)round(pixel_size), dpi, features);
@@ -1142,7 +1186,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
     uint8_t *data = NULL;
 
     FT_Error err;
-    if ((err = FT_Load_Glyph(inst->face, index, inst->load_flags)) != 0) {
+    if ((err = FT_Load_Glyph(inst->face, index, inst->load_flags)) != FT_Err_Ok) {
         LOG_ERR("%s: failed to load glyph #%d: %s",
                 inst->path, index, ft_error_string(err));
         goto err;
@@ -1184,6 +1228,14 @@ glyph_for_index(const struct instance *inst, uint32_t index,
         bgr = false;
     }
 
+#if FREETYPE_MAJOR >= 3 || (FREETYPE_MAJOR >= 2 && FREETYPE_MINOR >= 12)
+    if (inst->face->glyph->format == FT_GLYPH_FORMAT_SVG) {
+        /* Up to, and include, 2.12.1, FreeType rejects everything
+         * else with “bad argument” */
+        render_flags = FT_RENDER_MODE_NORMAL;
+    }
+#endif
+
     /*
      * LCD filter is per library instance, hence we need to re-set it
      * every time. But only if we need it, and only if we _can_, to
@@ -1199,7 +1251,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
 
         FT_Error err = FT_Library_SetLcdFilter(ft_lib, inst->lcd_filter);
 
-        if (err != 0) {
+        if (err != FT_Err_Ok) {
             LOG_ERR("failed to set LCD filter: %s", ft_error_string(err));
             mtx_unlock(&ft_lock);
             goto err;
@@ -1207,7 +1259,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
     }
 
     if (inst->face->glyph->format != FT_GLYPH_FORMAT_BITMAP) {
-        if ((err = FT_Render_Glyph(inst->face->glyph, render_flags)) != 0) {
+        if ((err = FT_Render_Glyph(inst->face->glyph, render_flags)) != FT_Err_Ok) {
             LOG_ERR("%s: failed to render glyph: %s",
                     inst->path, ft_error_string(err));
             if (unlock_ft_lock)
@@ -2218,6 +2270,7 @@ fcft_rasterize_grapheme_utf32(struct fcft_font *_font,
     return &grapheme->public;
 
 err:
+    hb_buffer_clear_contents(inst->hb_buf);
     for (size_t i = 0; i < glyph_idx; i++)
         glyph_destroy(grapheme->public.glyphs[i]);
     free(grapheme->public.glyphs);
@@ -2665,7 +2718,7 @@ fcft_kerning(struct fcft_font *_font, uint32_t left, uint32_t right,
     FT_Error err = FT_Get_Kerning(
         primary->face, left_idx, right_idx, FT_KERNING_DEFAULT, &kerning);
 
-    if (err != 0) {
+    if (err != FT_Err_Ok) {
         LOG_WARN("%s: failed to get kerning for %lc -> %lc: %s",
                  primary->path, (int)left, (int)right, ft_error_string(err));
         goto err;
